/*++

Copyright (c) 1996  Microsoft Corporation

Module Name:

    main.c

Abstract:

    This is the initialization file for the sample Ndis 4.0 Intermediate Driver. This sample is
	designed to layer in between the installed transports and the installed Miniports. Currently.
	this sample doesn't do anything except receive packets, and pass them on.
	The upper edge of this sample is a Miniport interface, the lower edge looks like a transport
Author:

    Jim Mateer 
   
   Environment:

    Kernel Mode

Revision History:

         Based largely on the Packet Scheduler progtam written by  Charlie Wickham
           
--*/

#include "ImSamp.h"

#pragma hdrstop

//
// number of characters that are appended to the RegistryPath when constructing
// the miniport device name
//

#define MPNAME_EXTENSION_SIZE   ( 3 * sizeof(WCHAR))



// Define some memory tags. These will help track down any allocation errors while debugging


ULONG NdisRequestTag =  'qerN';
ULONG AdapterTag = 'padA';
ULONG PacketContextTag = 'tkaP';
ULONG ResidualBufferTag = 'bseR';
ULONG LookaheadBufferTag = 'hakL';
ULONG MediaAreaTag = 'taeM';
ULONG ClientVCTag  = 'tcvC';

// and some DbgPrint controls
ULONG DbgTraceLevel;
ULONG DbgTraceMask;



NTSTATUS
DriverEntry(
    IN PDRIVER_OBJECT DriverObject,
    IN PUNICODE_STRING RegistryPath
    );

STATIC NDIS_STATUS
InitializeNdisWrapper(
    IN PDRIVER_OBJECT DriverObject,
    IN  PUNICODE_STRING RegistryPath
    );

STATIC NDIS_STATUS
DoMiniportInit(
    IN PDRIVER_OBJECT DriverObject,
    IN  PUNICODE_STRING RegistryPath
    );

STATIC NDIS_STATUS
DoProtocolInit(
    IN PDRIVER_OBJECT DriverObject,
    IN  PUNICODE_STRING RegistryPath
    );

STATIC VOID
InitializationCleanup(
    ULONG ShutdownMask
    );


// Mark the initialization files as pageable
// To conserve system resources 

#pragma NDIS_INIT_FUNCTION(DriverEntry)
#pragma NDIS_INIT_FUNCTION(InitializeNdisWrapper)
#pragma NDIS_INIT_FUNCTION(DoProtocolInit)
#pragma NDIS_INIT_FUNCTION(DoMiniportInit)
#pragma NDIS_INIT_FUNCTION(InitializationCleanup)




NTSTATUS
DriverEntry(
    IN PDRIVER_OBJECT DriverObject,
    IN PUNICODE_STRING RegistryPath
    )

/*++

Routine Description:

    This is the NT OS specific driver entry point.  It kicks off initialization
    for the driver.  Currently, it is not PnP aware. Return from this routine
    only after protocol registration, layered miniport registration, and both
    miniport and higher layer protocol initialization has been done.

Arguments:

    DriverObject - NT OS specific Object
    RegistryPath - NT OS specific pointer to registry location for ImSamp

Return Values:

    STATUS_SUCCESS
    STATUS_FAILURE

--*/
{
    NDIS_STATUS Status;
    ULONG FuncIndex;
    ULONG ErrorLogData[2];
    ULONG InitShutdownMask;
    PWCHAR EventLogString = IMDriverName.Buffer;
    PVOID DumpData;
    ULONG MaxTimerRes;
    ULONG MinTimerRes;
    ULONG CurrentTimerRes;
    ULONG ActualTimerRes;

#ifdef PACKET_POOL_LOG
    NDIS_PHYSICAL_ADDRESS HighAddress = NDIS_PHYSICAL_ADDRESS_CONST( -1, -1 );
#endif



   //
    // announce the version
    //

    KdPrint(( VersionHerald, VersionNumber, VersionTimestamp ));

    //
    // store a copy of our driver object. Used by NdisWriteEventLogEntry
    //

    IMDriverObject = DriverObject;

    //
    // initialize global data and ndis request lookaside list
    // 

      InitializeListHead( &AdapterList );
      NdisAllocateSpinLock( &AdapterListLock );
    //
    // init the LLists for NdisRequest.
    // Look at ExInitializeNPagedLookasideList(...) for documentation
    NdisInitializeNPagedLookasideList(&NdisRequestLL,
                                      NULL,
                                      NULL,
                                      0,
                                      sizeof( IM_NDIS_REQUEST ),
                                      NdisRequestTag,
                                      (USHORT)ConfigData.NdisRequestBufferLimit);


    //
    // Lookaside list depth is automatically managed by the executive. The depth arg
    // (as of 1/5/97) is ignored.
    //



    //
    // initialize the wrapper with NDIS
    //

    Status = InitializeNdisWrapper( DriverObject, RegistryPath );

    if ( !NT_SUCCESS( Status )) {

        KdPrint(("InitializeNdisWrapper - Status: 0x%x\n", Status ));

        NdisWriteErrorLogEntry(IMDriverObject,
                               EVENT_TRANSPORT_REGISTER_FAILED,
                               IM_ERROR_INIT_NDIS_WRAPPER,
                               1,
                               &EventLogString,
                               0,
                               NULL);

        goto DriverEntryError;
    }

    InitShutdownMask = SHUTDOWN_TERMINATE_WRAPPER;

    //
    // get driver wide configuration data from the registry
    //

    Status = ConfigureDriver( RegistryPath, &ConfigData ); // in config.c

    if ( !NT_SUCCESS( Status )) {

        KdPrint(("ConfigureDriver - Status: 0x%x\n", Status ));
        goto DriverEntryError;
    }

#if DBG

    DbgTraceLevel = ConfigData.DebugLevel;
    DbgTraceMask = ConfigData.DebugMask;
#endif

    //
    // do Protocol registration    //

    Status = DoProtocolInit( DriverObject, RegistryPath );

    if ( !NT_SUCCESS( Status )) {

        ImDbgOut(DBG_CRITICAL_ERROR,
                 DBG_INIT,
                 ("DoProtocolInit Failed! Status: 0x%x\n", Status));

        DumpData = &Status;
        NdisWriteErrorLogEntry(IMDriverObject,
                               EVENT_TRANSPORT_REGISTER_FAILED,
                               IM_ERROR_PROTOCOL_INIT,
                               1,
                               &EventLogString,
                               sizeof( Status ),
                               DumpData);

        goto DriverEntryError;
    }

    InitShutdownMask |= SHUTDOWN_DEREGISTER_PROTOCOL;


    //
    // Initialize as a Miniport driver to the transports. no shutdown mask value
    // since there is no way to dereg a miniport...
    //

    Status = DoMiniportInit( DriverObject, RegistryPath );

    if ( !NT_SUCCESS( Status )) {

        ImDbgOut(DBG_CRITICAL_ERROR,
                 DBG_INIT,
                 ("DoMiniportInit Failed! Status: 0x%x\n", Status));

        DumpData = &Status;
        NdisWriteErrorLogEntry(IMDriverObject,
                               (ULONG)IM_EVENT_MINIPORT_REGISTER_FAILED,
                               0,
                               1,
                               &EventLogString,
                               sizeof( Status ),
                               DumpData);

        goto DriverEntryError;
    }


  



    Status = WDMInitialize( DriverObject, &InitShutdownMask );

    if ( !NT_SUCCESS( Status )) {

        ImDbgOut(DBG_CRITICAL_ERROR,
                 DBG_INIT,
                 ("WDMInitialize Failed! Status: 0x%x\n", Status));

        goto DriverEntryError;
    }

#ifdef PACKET_POOL_LOG
    NdisAllocateMemory( &PktPoolLog, sizeof(PKTPOOLLOG) * PPLogSize, 0, HighAddress );
#endif


    return (STATUS_SUCCESS);

    //
    // An error occured so we need to cleanup things
    //

DriverEntryError:
    InitializationCleanup( InitShutdownMask );

    return (STATUS_UNSUCCESSFUL);

} // DriverEntry





STATIC NDIS_STATUS
InitializeNdisWrapper(
    IN PDRIVER_OBJECT DriverObject,
    IN  PUNICODE_STRING RegistryPath
    )

/*++

Routine Description:

    Initialize the Ndis wrapper for both the miniport and protocol sections. Since
    the name in the registry path is the Protocol key (ImSamp), 'MP' is appended
    onto the end to initialize the wrapper for the miniport side of the IM

Arguments:

    DriverObject - pointer to NT driver object
    RegistryPath - pointer to path to driver params in registry 

Return Values:

    NDIS_STATUS_SUCCESS
    NDIS_STATUS_BAD_VERSION
    NDIS_STATUS_FAILURE

--*/

{
    NDIS_STATUS Status;
    USHORT MPDeviceNameLength;
    NDIS_PHYSICAL_ADDRESS HighAddress = NDIS_PHYSICAL_ADDRESS_CONST( -1, -1 );
    ULONG i;
    NDIS_STRING MPRegistryPath;
    PWCHAR RegistryPathBuffer;


    NdisMInitializeWrapper( &MPWrapperHandle, DriverObject, RegistryPath, NULL );

    return NDIS_STATUS_SUCCESS;
} // InitializeNdisWrapper







STATIC NDIS_STATUS
DoMiniportInit(
    IN PDRIVER_OBJECT DriverObject,
    IN  PUNICODE_STRING RegistryPath
    )

/*++

Routine Name:

    DoMiniportInit

Routine Description:

    This routines registers ImSamp as a Miniport driver with the NDIS wrapper.

Arguments:

    None

Return Values:

    NDIS_STATUS_SUCCESS
    NDIS_STATUS_BAD_VERSION
    NDIS_STATUS_FAILURE

--*/

{
    NDIS_STATUS Status;
    NDIS_MINIPORT_CHARACTERISTICS MiniportChars;

/////////////////////////////////////////////////////////////////////////////////
	PNDIS_MINIPORT_BLOCK Testing;
    NDIS_PHYSICAL_ADDRESS HighAddress = NDIS_PHYSICAL_ADDRESS_CONST( -1, -1 );
    NdisAllocateMemory( &Testing, sizeof(NDIS_MINIPORT_BLOCK), 0, HighAddress );
  //////////////////////////////////////////////////////////////////////////////
    MiniportChars.MajorNdisVersion = 4;
    MiniportChars.MinorNdisVersion = 0;

    MiniportChars.Reserved = 0;
    MiniportChars.HaltHandler = MPHalt;							  
    MiniportChars.InitializeHandler = MPInitialize;				  
    MiniportChars.QueryInformationHandler = MPQueryInformation;	  
    MiniportChars.ResetHandler = MPReset;                         
    MiniportChars.SetInformationHandler = MPSetInformation;		  
    MiniportChars.TransferDataHandler = MPTransferData;			  

    //
    // Unused handlers
    //

    MiniportChars.ReconfigureHandler = NULL;	                  
    MiniportChars.DisableInterruptHandler = NULL;				  
    MiniportChars.EnableInterruptHandler = NULL;				  
    MiniportChars.HandleInterruptHandler = NULL;				  
    MiniportChars.ISRHandler = NULL;							  
    MiniportChars.CheckForHangHandler = NULL;					  

    //
    // Ndis 4.0 handlers. No regular send routine since we have a
    // SendPackets handler.
    //

    MiniportChars.ReturnPacketHandler = MPReturnPacket;	             
    MiniportChars.SendPacketsHandler = MPSendPackets;				 
    MiniportChars.AllocateCompleteHandler = NULL;					 
    MiniportChars.SendHandler = NULL;								 


    Status = NdisIMRegisterLayeredMiniport(MPWrapperHandle,			
                                           &MiniportChars,
                                           sizeof(MiniportChars),
                                            &LMDriverHandle);

    return (Status);
} // DoMiniportInit


STATIC NDIS_STATUS
DoProtocolInit(
    IN PDRIVER_OBJECT DriverObject,
    IN  PUNICODE_STRING RegistryPath
    )
/*++

Routine Name:

    DoProtocolInit

Routine Description:

    This function registers the IM as a protocol 
Arguments:

    RegistryPath - pointer to our key in the registry

Return Values:

    NDIS_STATUS_BAD_CHARACTERISTICS
    NDIS_STATUS_BAD_VERSION
    NDIS_STATUS_RESOURCES
    NDIS_STATUS_SUCCESS

--*/
{
    NDIS_PROTOCOL_CHARACTERISTICS ProtocolChars;
    NDIS_STATUS Status;
    NDIS_STRING IMName = NDIS_STRING_CONST( "IM");
   
    //
    // register the protocol section
    //
	NdisZeroMemory(&ProtocolChars, sizeof(NDIS_PROTOCOL_CHARACTERISTICS));
    ProtocolChars.Name.Length = IMName.Length;
    ProtocolChars.Name.Buffer = (PVOID)IMName.Buffer;

    ProtocolChars.MajorNdisVersion = 4;
    ProtocolChars.MinorNdisVersion = 0;

    ProtocolChars.OpenAdapterCompleteHandler = LowerMPOpenAdapterComplete;	  
    ProtocolChars.CloseAdapterCompleteHandler = LowerMPCloseAdapterComplete;  
    ProtocolChars.SendCompleteHandler = CLSendComplete;						  
    ProtocolChars.TransferDataCompleteHandler = CLTransferDataComplete;		  
    ProtocolChars.ResetCompleteHandler = CLResetComplete;                     
    ProtocolChars.RequestCompleteHandler = CLRequestComplete;                 
    ProtocolChars.ReceiveHandler = CLReceiveIndication;                        
    ProtocolChars.ReceiveCompleteHandler = CLReceiveComplete;                 
    ProtocolChars.StatusHandler = CLStatusIndication;	                      
    ProtocolChars.StatusCompleteHandler = CLStatusIndicationComplete;          
    ProtocolChars.ReceivePacketHandler = CLReceivePacket;                     
    ProtocolChars.BindAdapterHandler = BindToLowerMP;	                     
    ProtocolChars.UnbindAdapterHandler = UnbindFromLowerMP;                    
    ProtocolChars.UnloadHandler = CLUnloadProtocol;	                          




    NdisRegisterProtocol(&Status,
                         &ClientProtocolHandle,
                         &ProtocolChars,
                         sizeof(NDIS_PROTOCOL_CHARACTERISTICS) + ProtocolChars.Name.Length);

    if ( !NT_SUCCESS( Status )) {

        ImDbgOut(DBG_CRITICAL_ERROR, DBG_PROTOCOL | DBG_INIT,
                 ("DoProtocolInit: couldn't register client handlers %08X\n", Status ));

    }

    return Status;

}



STATIC VOID
InitializationCleanup(
    ULONG ShutdownMask
    )

/*++

Routine Description:

    This routine is responsible for cleaning up all allocated resources during
    initialization

Arguments:

    None

Return Values:

    None

--*/

{
    NDIS_STATUS Status;

    //
    // Deregister the protocol; we should have no references that would cause
    // this to pend
    //

    if ( ShutdownMask & SHUTDOWN_DEREGISTER_PROTOCOL ) {

        if ( ClientProtocolHandle ) {

            NdisDeregisterProtocol( &Status, ClientProtocolHandle );

            if ( Status == NDIS_STATUS_PENDING ) {

                ImDbgOut(DBG_CRITICAL_ERROR, DBG_INIT, ("Client DeregProto failed - %08X\n", Status));
            }
        }
	}
     //
    // Terminate the wrapper
    //

    if ( ShutdownMask & SHUTDOWN_TERMINATE_WRAPPER ) {

        NdisTerminateWrapper( MPWrapperHandle, NULL );
    }

    //
    // Delete our IO device and symbolic link
    //

    WDMCleanup( ShutdownMask );

    //
    // free the lookaside list resources
    //

    NdisDeleteNPagedLookasideList( &NdisRequestLL );

    //
    // Free our global locks
    //

    NdisFreeSpinLock( &AdapterListLock );
    NdisFreeSpinLock( &PSAListLock );

} // InitializationCleanup

/* end main.c */

